/***************************************************************************/
/* 		This code is part of WWW grabber called pavuk		   */
/*		Copyright (c) 1997 - 2000 Stefan Ondrejicka		   */
/*		(ondrej@idata.sk)					   */
/*		Distributed under GPL 2 or later			   */
/***************************************************************************/

#include "config.h"

/****************************************************************/
/* based on document "NTLM Athentication Scheme for HTTP"       */
/* http://www.innovation.ch/java/ntlm.html                      */
/* by Ronald Tschaler <ronald@innovation.ch>                    */
/****************************************************************/

#ifdef ENABLE_NTLM

#include <string.h>
#include <ctype.h>

#include "base64.h"
#include "tools.h"
#include "mime.h"
#include "doc.h"
#include "http.h"
#include "tools.h"
#include "abstract.h"
#include "authinfo.h"
#include "errcode.h"
#include "ntlm_auth.h"


#ifdef OPENSSL
#include <openssl/des.h>
#else
#include <des.h>
#endif

#ifdef HAVE_OPENSSL_MD4
#include <openssl/md4.h>

#define _MD4Init	MD4_Init
#define _MD4Update	MD4_Update
#define _MD4Final	MD4_Final

#else
#include "md4c.h"

#define _MD4Init	MD4Init
#define _MD4Update	MD4Update
#define _MD4Final	MD4Final
#endif

#ifdef WORDS_BIGENDIAN
static short ntlm_do_litle_endian_short(v)
short v;
{
	short sh;

	((char *)&sh)[0] = v && 0xFF;
	((char *)&sh)[1] = v >> 8;

	return sh;
}

static short ntlm_get_litle_endian_short(v)
short v;
{
	short sh;

	sh = ((char *)&v)[0] + (((char *)&v)[1] << 8)

	return sh;
}
#else
#define ntlm_do_litle_endian_short(v) (v)
#define ntlm_get_litle_endian_short(v) (v)
#endif

char *ntlm_get_t1msg_str(domain, host)
char *domain;
char *host;
{
	ntlm_type_1_msg_t msg;
	int msg_size;

	memset(&msg, '\0', sizeof(msg));
	strcpy(msg.protocol, "NTLMSSP");
	msg.type = 0x01;
	msg.flags = ntlm_do_litle_endian_short(0xb203);
	msg.dom_len2 = msg.dom_len =
		ntlm_do_litle_endian_short(strlen(domain));
	msg.host_len2 = msg.host_len =
		ntlm_do_litle_endian_short(strlen(host));
	msg.host_off = ntlm_do_litle_endian_short(32);
	msg.dom_off = msg.host_off + msg.host_len;

	strcpy(((char *)&msg) + msg.host_off, host);
	upperstr(((char *)&msg) + msg.host_off);
	strcpy(((char *)&msg) + msg.dom_off, domain);
	upperstr(((char *)&msg) + msg.dom_off);

	msg_size = 32 + msg.dom_len + msg.host_len;

	return base64_encode_data((char *)&msg, msg_size);
}

char *ntlm_get_nonce(t2msg)
char *t2msg;
{
	ntlm_type_2_msg_t *msg = NULL;
	char *rv;
	int st;

	st = base64_decode_data(t2msg, (char **)&msg);

	if (st < 0 || st != sizeof(ntlm_type_2_msg_t) ||
		strcmp(msg->protocol, "NTLMSSP") ||
		msg->type != 0x02)
	{
		rv = NULL;
	}
	else
	{
		rv = _malloc(9);
		memcpy(rv, msg->nonce, 8);
		rv[8] = '\0';
	}

	_free(msg);

	return rv;
}

static void ntlm_copy_str_to_wcstr(wcstr, str)
char *wcstr;
char *str;
{
	while (*str)
	{
		wcstr[0] = *str;
		wcstr[1] = '\0';

		wcstr += 2;
		str ++;
	}
}

static void ntlm_copy_str_to_uwcstr(wcstr, str)
char *wcstr;
char *str;
{
	while (*str)
	{
		wcstr[0] = toupper(*str);
		wcstr[1] = '\0';

		wcstr += 2;
		str ++;
	}
}


/*
* turns a 56 bit key into the 64 bit, odd parity key and sets the key.
* The key schedule ks is also set.
*/
static void ntlm_setup_des_key(key_56, ks)
unsigned char *key_56;
des_key_schedule ks;
{
	des_cblock key;

	key[0] = key_56[0];
	key[1] = ((key_56[0] << 7) & 0xFF) | (key_56[1] >> 1);
	key[2] = ((key_56[1] << 6) & 0xFF) | (key_56[2] >> 2);
	key[3] = ((key_56[2] << 5) & 0xFF) | (key_56[3] >> 3);
	key[4] = ((key_56[3] << 4) & 0xFF) | (key_56[4] >> 4);
	key[5] = ((key_56[4] << 3) & 0xFF) | (key_56[5] >> 5);
	key[6] = ((key_56[5] << 2) & 0xFF) | (key_56[6] >> 6);
	key[7] =  (key_56[6] << 1) & 0xFF;

	des_set_odd_parity(key);
	des_set_key(key, ks);
}

/*
 * takes a 21 byte array and treats it as 3 56-bit DES keys. The
 * 8 byte plaintext is encrypted with each key and the resulting 24
 * bytes are stored in the results array.
 */
static void ntlm_calc_resp(keys, plaintext, results)
unsigned char *keys;
unsigned char *plaintext;
unsigned char *results;
{
	des_key_schedule ks;

	ntlm_setup_des_key(keys, ks);
	des_ecb_encrypt(plaintext, results, ks, DES_ENCRYPT);

	ntlm_setup_des_key(keys+7, ks);
	des_ecb_encrypt(plaintext, (results+8), ks, DES_ENCRYPT);

	ntlm_setup_des_key(keys+14, ks);
	des_ecb_encrypt(plaintext, (results+16), ks, DES_ENCRYPT);
}

void ntlm_gen_lm_resp(pass, nonce, buf)
char *pass;
char *nonce;
char *buf;
{
	unsigned char magic[] = {0x4B, 0x47, 0x53, 0x21, 0x40, 0x23, 0x24, 0x25};
	unsigned char lm_hpw[21];
	char lm_pw[14];
	des_key_schedule ks;

	memset(lm_pw, '\0', sizeof(lm_pw));
	strncpy(lm_pw, pass, 14);
	upperstr(lm_pw);

	ntlm_setup_des_key(lm_pw, ks);
	des_ecb_encrypt(magic, lm_hpw, ks, DES_ENCRYPT);

	ntlm_setup_des_key(lm_pw+7, ks);
	des_ecb_encrypt(magic, lm_hpw+8, ks, DES_ENCRYPT);

	memset(lm_hpw+16, '\0', 5);

	ntlm_calc_resp(lm_hpw, nonce, buf);
}

void ntlm_gen_nt_resp(pass, nonce, buf)
char *pass;
char *nonce;
char *buf;
{
        MD4_CTX md4ctx;
	char *nt_pw;
	unsigned char nt_hpw[21];
	int len = strlen(pass);

	nt_pw = _malloc(2*len);
	ntlm_copy_str_to_wcstr(nt_pw, pass);

	_MD4Init(&md4ctx);
	_MD4Update(&md4ctx, nt_pw, 2*len);
	_MD4Final(nt_hpw , &md4ctx);

	memset(nt_hpw+16, '\0', 5);

	ntlm_calc_resp(nt_hpw, nonce, buf);

	_free(nt_pw);
}

char *ntlm_get_t3msg_str(domain, host, user, pass, nonce)
char *domain;
char *host;
char *user;
char *pass;
char *nonce;
{
	ntlm_type_3_msg_t msg;
	int msg_size;

	memset(&msg, '\0', sizeof(msg));
	strcpy(msg.protocol, "NTLMSSP");
	msg.type = 0x03;
	msg.flags = 0x8201;

	msg.dom_len = msg.dom_len2 =
		ntlm_do_litle_endian_short(2 * strlen(domain));
	msg.dom_off = ntlm_do_litle_endian_short(64);
	msg.user_len = msg.user_len2 = 
		ntlm_do_litle_endian_short(2 * strlen(user));
	msg.user_off = ntlm_do_litle_endian_short(msg.dom_off + msg.dom_len);
	msg.host_len = msg.host_len2 =
		ntlm_do_litle_endian_short(2 * strlen(host));
	msg.host_off = ntlm_do_litle_endian_short(msg.user_off + msg.user_len);

	msg.lm_resp_len = msg.lm_resp_len2 = ntlm_do_litle_endian_short(0x18);
	msg.lm_resp_off =
		ntlm_do_litle_endian_short(msg.host_off + msg.host_len);
	msg.nt_resp_len = msg.nt_resp_len2 = ntlm_do_litle_endian_short(0x18);
	msg.nt_resp_off =
		ntlm_do_litle_endian_short(msg.lm_resp_off + msg.lm_resp_len);

	ntlm_copy_str_to_uwcstr(((char *)&msg) + msg.dom_off, domain);
	ntlm_copy_str_to_wcstr(((char *)&msg) + msg.user_off, user);
	ntlm_copy_str_to_uwcstr(((char *)&msg) + msg.host_off, host);

	ntlm_gen_lm_resp(pass, nonce, ((char *)&msg) + msg.lm_resp_off);
	ntlm_gen_nt_resp(pass, nonce, ((char *)&msg) + msg.nt_resp_off);

	msg_size = 64 + msg.dom_len + msg.user_len + msg.host_len + msg.lm_resp_len + msg.nt_resp_len;

	msg.msg_len = ntlm_do_litle_endian_short(msg_size);

	return base64_encode_data((char *)&msg, msg_size);
}

/*
 * -1 - failure before sending auth data - http_process_response() can
 *      continue safely
 *  0 - OK - http_process_response() must return immediatly
 *  1 - failure after sending auth data - http_process_response() must
 *      return immediatly
 */
int ntlm_negotiate_connection(docp, authtag)
doc *docp;
char *authtag;
{
	char *ntlm_msg, *nonce;
	http_response *resp;
	char *user, *pass, *domain, *host;
	authinfo *ai;
	char shost[128];
	int l;

	if (docp->num_auth)
		return -1;

	docp->num_auth ++;

	xprintf(1, gettext("Trying to do NTLM authorization\n"));

	/*** supported only when persistant connections are  ***/
	/*** supported by server and we are talking HTTP/1.1 ***/
	if (!docp->is_persistent)
	{
		xprintf(1, gettext("NTLM authorization supported only on persistent connections!\n"));
		docp->errcode = ERR_HTTP_AUTH_NTLM;
		return -1;
	}

	/*** get authorization infos (all mandatory) ***/
	ai = authinfo_match_entry(docp->doc_url->type,
		url_get_site(docp->doc_url), url_get_port(docp->doc_url),
		url_get_path(docp->doc_url), NULL);
	domain = (ai && ai->ntlm_domain) ?
			ai->ntlm_domain : priv_cfg.auth_ntlm_domain;

	l = strcspn(cfg.local_host, ".");
	strncpy(shost, cfg.local_host, l);
	shost[l] = '\0';
	host = shost;

	user = url_get_user(docp->doc_url, NULL);
        pass = url_get_pass(docp->doc_url, NULL);

	if (!domain || !user || !pass || !host)
	{
		xprintf(1, gettext("Not enough data for NTLM authorization!\n"));
		xprintf(1, gettext("Missing:\n"));
		if (!domain) xprintf(1, gettext("       domain\n"));
		if (!user)   xprintf(1, gettext("       username\n"));
		if (!pass)   xprintf(1, gettext("       password\n"));
		if (!host)   xprintf(1, gettext("       local hostname\n"));
		return -1;
	}

	/*** read body of 401 response ***/
	if (http_throw_message_body(docp))
	{
		docp->is_persistent = FALSE;
		abs_close_socket(docp, FALSE);
		return 1;
	}

	/*** prepare and send NTLM msg type 1 to ***/
	/*** start NTLM negotiation with server  ***/
	ntlm_msg = ntlm_get_t1msg_str(domain, host);

	docp->additional_headers = tl_str_concat(NULL, "Authorization: NTLM ",
			ntlm_msg, "\r\n", NULL);
	_free(ntlm_msg);
	_free(docp->mime);
	_free(docp->type_str);

	l = http_repeat_request(docp);

	_free(docp->additional_headers);

	if (l)
		return -1;

	/*** now we expect 401 server response with ***/
	/*** NTLM msg type 2 which contains nonce   ***/
	resp = http_get_response_info(docp->mime);

	if (!resp)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		if (!docp->errcode)
			docp->errcode = ERR_HTTP_AUTH_NTLM;
		return -1;
	}

	if (resp->ret_code != 401)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		docp->errcode = ERR_HTTP_AUTH_NTLM;
		http_response_free(resp);
		return -1;
	}
	http_response_free(resp);

	if (!docp->is_persistent)
	{
		xprintf(1, gettext("NTLM authorization supported only on persistent connections!\n"));
		docp->errcode = ERR_HTTP_AUTH_NTLM;
		return -1;
	}

	if (http_throw_message_body(docp))
	{
		docp->is_persistent = FALSE;
		abs_close_socket(docp, FALSE);
		return 1;
	}

	authtag = get_mime_param_val_str("WWW-Authenticate: NTLM ", docp->mime);

	if (!authtag)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		docp->errcode = ERR_HTTP_AUTH_NTLM;
		return -1;
	}

	nonce = ntlm_get_nonce(authtag);
	_free(authtag);

	if (!nonce)
	{
		xprintf(1, gettext("Failed NTLM nonce negotiation\n"));
		docp->errcode = ERR_HTTP_AUTH_NTLM;
		return -1;
	}

	/*** final step of NTLM authorization - prepare    ***/
	/*** and send NTLM msg type 3 with all auth. infos ***/
	ntlm_msg = ntlm_get_t3msg_str(domain, host, user, pass, nonce);
	docp->additional_headers = tl_str_concat(NULL, "Authorization: NTLM ",
			ntlm_msg, "\r\n", NULL);
	_free(ntlm_msg);
	_free(nonce);
	_free(docp->mime);
	_free(docp->type_str);

	l = http_repeat_request(docp);

	_free(docp->additional_headers);

	if (l)
		return -1;

	return 0;
}

int ntlm_negotiate_proxy_connection(docp, authtag)
doc *docp;
char *authtag;
{
	char *ntlm_msg, *nonce;
	http_response *resp;
	char *user, *pass, *domain, *host;
	authinfo *ai;
	char shost[128];
	int l;

	if (docp->num_proxy_auth)
		return -1;

	if (!docp->http_proxy)
		return -1;

	xprintf(1, gettext("Trying to do proxy NTLM authorization\n"));

	docp->num_proxy_auth ++;

	docp->errcode = ERR_HTTP_PROAUTH_NTLM;

	/*** supported only when persistant connections are  ***/
	/*** supported by server and we are talking HTTP/1.1 ***/
	if (!docp->is_persistent)
	{
		xprintf(1, gettext("NTLM authorization supported only on persistent connections!\n"));
		return -1;
	}

	/*** get authorization infos (all mandatory) ***/
	ai = authinfo_match_entry(docp->doc_url->type,
		docp->http_proxy, docp->http_proxy_port,
		NULL, NULL);
	domain = (ai && ai->ntlm_domain) ?
			ai->ntlm_domain : priv_cfg.auth_proxy_ntlm_domain;

	l = strcspn(cfg.local_host, ".");
	strncpy(shost, cfg.local_host, l);
	shost[l] = '\0';
	host = shost;

	user = priv_cfg.http_proxy_user;
        pass = priv_cfg.http_proxy_pass;

	if (!domain || !user || !pass || !host)
	{
		xprintf(1, gettext("Not enough data for NTLM authorization!\n"));
		xprintf(1, gettext("Missing:\n"));
		if (!domain) xprintf(1, gettext("       domain\n"));
		if (!user)   xprintf(1, gettext("       username\n"));
		if (!pass)   xprintf(1, gettext("       password\n"));
		if (!host)   xprintf(1, gettext("       local hostname\n"));
		return -1;
	}

	/*** read body of 407 response ***/
	if (http_throw_message_body(docp))
	{
		docp->is_persistent = FALSE;
		abs_close_socket(docp, FALSE);
		return 1;
	}

	/*** prepare and send NTLM msg type 1 to ***/
	/*** start NTLM negotiation with server  ***/
	ntlm_msg = ntlm_get_t1msg_str(domain, host);

	docp->additional_headers = tl_str_concat(NULL, "Proxy-Authorization: NTLM ",
			ntlm_msg, "\r\n", NULL);

	_free(ntlm_msg);
	_free(docp->mime);
	_free(docp->type_str);

	l = http_repeat_request(docp);
	_free(docp->additional_headers);

	if (l)
		return -1;

	/*** now we expect 407 server response with ***/
	/*** NTLM msg type 2 which contains nonce   ***/
	resp = http_get_response_info(docp->mime);

	if (!resp)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		if (!docp->errcode)
			docp->errcode = ERR_HTTP_PROAUTH_NTLM;
		return -1;
	}

	if (resp->ret_code != 407)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		docp->errcode = ERR_HTTP_PROAUTH_NTLM;
		http_response_free(resp);
		return -1;
	}
	http_response_free(resp);

	if (!docp->is_persistent)
	{
		xprintf(1, gettext("NTLM authorization supported only on persistent connections!\n"));
		docp->errcode = ERR_HTTP_PROAUTH_NTLM;
		return -1;
	}

	if (http_throw_message_body(docp))
	{
		docp->is_persistent = FALSE;
		abs_close_socket(docp, FALSE);
		return 1;
	}

	authtag = get_mime_param_val_str("Proxy-Authenticate: NTLM ", docp->mime);

	if (!authtag)
	{
		xprintf(1, gettext("Got unexpected response\n"));
		docp->errcode = ERR_HTTP_PROAUTH_NTLM;
		return -1;
	}

	nonce = ntlm_get_nonce(authtag);
	_free(authtag);

	if (!nonce)
	{
		xprintf(1, gettext("Failed NTLM proxy nonce negotiation\n"));
		docp->errcode = ERR_HTTP_PROAUTH_NTLM;
		return -1;
	}

	/*** final step of NTLM authorization - prepare    ***/
	/*** and send NTLM msg type 3 with all auth. infos ***/
	ntlm_msg = ntlm_get_t3msg_str(domain, host, user, pass, nonce);
	docp->additional_headers = tl_str_concat(NULL, "Proxy-Authorization: NTLM ",
			ntlm_msg, "\r\n", NULL);
	_free(ntlm_msg);
	_free(nonce);
	_free(docp->mime);
	_free(docp->type_str);

	l = http_repeat_request(docp);

	if (l)
		return -1;

	return 0;
}

#endif /* ENABLE_NTLM */
